package cds.simbad.tap;

import java.io.File;
import java.io.IOException;
import java.sql.ResultSet;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;

import cds.simbad.tap.cdsLogin.CDSLoginResource;
import cds.simbad.tap.formatter.SimbadFormatter;
import cds.util.Parameters;
import edu.sdsc.grid.io.irods.IRODSAccount;
import tap.ServiceConnection;
import tap.TAPFactory;
import tap.file.TAPFileManager;
import tap.file.iRodsTAPFileManager;
import tap.formatter.OutputFormat;
import tap.formatter.ResultSet2JsonFormatter;
import tap.formatter.ResultSet2SVFormatter;
import tap.formatter.ResultSet2TextFormatter;
import tap.formatter.ResultSet2VotableFormatter;
import tap.formatter.SVFormat;
import tap.log.DefaultTAPLog;
import tap.log.TAPLog;
import tap.metadata.TAPMetadata;
import uws.UWSException;
import uws.service.UserIdentifier;

/**
 * Description of the Simbad-TAP service.
 * 
 * @author Gr&eacute;gory Mantelet (CDS)
 * @version 06/2012
 */
public class SimbadServiceConnection implements ServiceConnection<ResultSet> {

	private final TAPFactory<ResultSet> factory;

	private final TAPLog logger;

	private final TAPFileManager fileManager;

	private final UserIdentifier userIdentifier;

	private final ArrayList<OutputFormat<ResultSet>> outputFormats;
	private final TAPMetadata tapSchema;

	private boolean available = true;
	private String availability = MSG_AVAILABLE;

	public final static String MSG_AVAILABLE = "Simbad TAP is accepting queries.";
	public final static String MSG_UNAVAILABLE = "Simbad TAP is temporarily unavailable.";

	/** in seconds */
	private int retentionPeriod = -1;
	private int anonymousRetentionPeriod = -1;

	/** in seconds */
	private int executionDuration = -1;

	private int outputLimit = -1;
	private LimitUnit outputLimitType = LimitUnit.rows;

	private int uploadLimit = -1;
	private LimitUnit uploadLimitType = LimitUnit.rows;

	//private String uploadDirectory = null;


	public SimbadServiceConnection() throws NullPointerException, IOException, UWSException {
		// Builds an iRods account:
		// Build the account to write result and backup files:
		IRODSAccount account = new IRODSAccount(Parameters.get("tapstore", "irods", "host"),
				Integer.parseInt(Parameters.get("tapstore", "irods", "port")),
				Parameters.get("tapstore", "irods", "userName"),
				Parameters.get("tapstore", "irods", "password"),
				Parameters.get("tapstore", "irods", "homeDirectory"),
				Parameters.get("tapstore", "irods", "zone"),
				Parameters.get("tapstore", "irods", "defaultStorageResource"));

		// Create the file manager:
		//fileManager = new LocalTAPFileManager(new File("/home/mantelet/Bureau/simbadTap"));
		fileManager = new iRodsTAPFileManager(account, new File(Parameters.get("tap", "tap", "logDirectory")), new File(Parameters.get("tap", "tap", "uploadDirectory")));

		// Create the logger:
		logger = new DefaultTAPLog(fileManager);

		// Create all the formatters:
		outputFormats = new ArrayList<OutputFormat<ResultSet>>(5);
		outputFormats.add(new SimbadFormatter(new ResultSet2VotableFormatter(this)));
		outputFormats.add(new SimbadFormatter(new ResultSet2JsonFormatter(this)));
		outputFormats.add(new SimbadFormatter(new ResultSet2SVFormatter(this, SVFormat.COMMA_SEPARATOR)));
		outputFormats.add(new SimbadFormatter(new ResultSet2SVFormatter(this, SVFormat.TAB_SEPARATOR)));
		outputFormats.add(new SimbadFormatter(new ResultSet2TextFormatter(this)));

		tapSchema = new SimbadTAPMetadata(logger);

		// Configure the destruction time:
		retentionPeriod = Integer.parseInt(Parameters.get("tap", "tap", "destructionTime"))*86400;
		anonymousRetentionPeriod = Integer.parseInt(Parameters.get("tap", "tap", "anonymousDestructionTime"))*86400;

		// Configure the executionDuration:
		executionDuration = Integer.parseInt(Parameters.get("tap", "tap", "executionDuration"))*60;

		// Configure the output limit:
		outputLimit = Integer.parseInt(Parameters.get("tap", "tap", "outputLimit"));

		// Configure the upload limit:
		uploadLimit = Integer.parseInt(Parameters.get("tap", "tap", "uploadLimit"));

		// // Get the upload directory:
		// uploadDirectory = Parameters.get("tap", "tap", "uploadDirectory");

		// Create the TAP factory:
		factory = new SimbadTAPFactory(this);

		// Create the User Identifier:
		userIdentifier = new CDSLoginResource(logger);
	}

	public String getProviderName() {
		return "Simbad";
	}

	public String getProviderDescription() {
		return "TAP Service of Simbad";
	}

	public boolean isAvailable() {
		return available;
	}

	public void setAvailable(final boolean available){
		this.available = available;
		availability = available ? MSG_AVAILABLE : MSG_UNAVAILABLE;
	}

	public String getAvailability() {
		return availability;
	}

	public void setAvailability(final String msg){
		availability = msg;
	}

	public final TAPMetadata getTAPMetadata() {
		return tapSchema;
	}

	public final Collection<String> getCoordinateSystems(){
		ArrayList<String> coordSys = new ArrayList<String>(2);
		coordSys.add("ICRS");
		coordSys.add("ICRS BARYCENTER");
		return coordSys;
	}

	public Iterator<OutputFormat<ResultSet>> getOutputFormats() {
		return outputFormats.iterator();
	}

	public OutputFormat<ResultSet> getOutputFormat(final String mimeOrAlias) {
		if (mimeOrAlias == null || mimeOrAlias.trim().isEmpty())
			return null;

		for(OutputFormat<ResultSet> f : outputFormats){
			if ((f.getMimeType() != null && f.getMimeType().equalsIgnoreCase(mimeOrAlias))
					|| (f.getShortMimeType() != null && f.getShortMimeType().equalsIgnoreCase(mimeOrAlias)))
				return f;
		}
		return null;
	}

	public int[] getExecutionDuration() {
		return new int[]{executionDuration, executionDuration};
	}

	/**
	 * Sets the maximum (and default) execution duration
	 * and then, updates the corresponding parameters file.
	 * 
	 * @param duration	Maximum execution duration (in minutes).
	 */
	public void setExecutionDuration(final int duration){
		if (duration > 0){
			executionDuration = duration*60;
			Parameters.getParameters("tap").update("tap", "executionDuration", duration+"", true);
		}
	}

	public int[] getOutputLimit() {
		return new int[]{-1, outputLimit};
	}

	/**
	 * Sets the maximum output limit
	 * and then, updates the corresponding parameters file.
	 * 
	 * @param limit	The maximum output limit (in rows).
	 */
	public void setOutputLimit(final int limit){
		if (limit > 0){
			outputLimit = limit;
			Parameters.getParameters("tap").update("tap", "outputLimit", limit+"", true);
		}
	}

	public LimitUnit[] getOutputLimitType() {
		return new LimitUnit[]{outputLimitType, outputLimitType};
	}

	public int[] getRetentionPeriod() {
		return new int[]{retentionPeriod, retentionPeriod};
	}

	public int getAnonymousRetentionPeriod(){
		return anonymousRetentionPeriod;
	}

	/**
	 * Sets the retention period (that's to say: how much time a job can stay on server side)
	 * and then, updates the corresponding parameters file.
	 * 
	 * @param period	Period during a job can stay on the server (in days).
	 */
	public void setRetentionPeriod(final int period){
		if (period > 0){
			retentionPeriod = period*86400;
			Parameters.getParameters("tap").update("tap", "destructionTime", period+"", true);
		}
	}

	/**
	 * Sets the retention period of the anonymous jobs (that's to say: how much time a job can stay on server side)
	 * and then, updates the corresponding parameters file.
	 * 
	 * @param period	Period during a job created by an anonymous can stay on the server (in days).
	 */
	public void setAnonymousRetentionPeriod(final int period){
		if (period > 0){
			anonymousRetentionPeriod = period*86400;
			Parameters.getParameters("tap").update("tap", "anonymousDestructionTime", period+"", true);
		}
	}

	public UserIdentifier getUserIdentifier(){
		return userIdentifier;
	}

	public boolean uploadEnabled(){
		return true;
	}

	public int[] getUploadLimit() {
		return new int[]{uploadLimit, uploadLimit};
	}

	public void setUploadLimit(final int limit){
		if (limit > 0){
			uploadLimit = limit;
			Parameters.getParameters("tap").update("tap", "uploadLimit", limit+"", true);
		}
	}

	public LimitUnit[] getUploadLimitType() {
		return new LimitUnit[]{uploadLimitType, uploadLimitType};
	}

	public int getMaxUploadSize() {
		return 1024*1024;
	}

	/*public String getUploadDirectory(){
		return uploadDirectory;
	}

	public void deleteResults(TAPJob<ResultSet> job) throws TAPException {
		if (job.getResult(Result.DEFAULT_RESULT_NAME) != null){
			IRODSFile file = ((IRodsResult)job.getResult(Result.DEFAULT_RESULT_NAME)).getFile();
			if (file != null){
				if (!file.delete())
					throw new TAPException("Can not delete the file \""+file+"\" associated with the result of "+job.getJobId()+" !");
			}
		}
	}*/

	public TAPFactory<ResultSet> getFactory() {
		return factory;
	}

	/*public void log(final LogType type, final String message){
		switch(type){
		case INFO:
			SimbadTapLog.info(message); break;
		case WARNING:
			SimbadTapLog.warning(message); break;
		case ERROR:
			SimbadTapLog.error(message); break;
		default:
			System.err.println(type+": "+message);
		}
	}*/

	/*public void log(final Throwable error){
		SimbadTapLog.error(error);
	}*/

	public TAPFileManager getFileManager() {
		return fileManager;
	}

	public TAPLog getLogger() {
		return logger;
	}

}
