package tap.file;

import java.io.File;
import java.io.IOException;

import edu.sdsc.grid.io.irods.IRODSAccount;
import tap.db.DBConnection;
import uws.UWSException;
import uws.job.user.JobOwner;
import uws.service.file.DefaultOwnerGroupIdentifier;
import uws.service.file.OwnerGroupIdentifier;
import uws.service.file.UWSFileManager;
import uws.service.file.iRodsUWSFileManager;

/**
 * <p>All UWS files and uploaded TAP files are stored in the local machine into the specified directory.</p>
 * <p>
 * 	The name of the log file, the result files and the backup files may be customized by overriding the following functions:
 * 	{@link #getLogFileName()}, {@link #getResultFileName(Result, AbstractJob)}, {@link #getBackupFileName(JobOwner)} and {@link #getBackupFileName()}.
 * </p>
 * <p>
 * 	By default, results and backups are grouped by owner/user and owners/users are grouped thanks to {@link DefaultOwnerGroupIdentifier}.
 * 	By using the appropriate constructor, you can change these default behaviors.
 * </p>
 * 
 * @author Gr&eacute;gory Mantelet (CDS)
 * @version 06/2012
 * 
 * @see iRodsUWSFileManager
 */
public class iRodsTAPFileManager extends iRodsUWSFileManager implements TAPFileManager {

	/** Default name of the DB activity log file. */
	public final static String DEFAULT_DB_ACTIVITY_LOG_FILE_NAME = "service_db_activity.log";

	/** Local directory in which all uploaded files will be kept until they are read or ignored (in this case, they will be deleted). */
	private final File uploadDirectory;


	/**
	 * <p>Builds a {@link TAPFileManager} which manages all UWS files in the home directory of the given IRODS account.</p>
	 * <p>
	 * 	There will be one directory for each owner ID and owner directories will be grouped
	 * 	thanks to {@link DefaultOwnerGroupIdentifier}.
	 * </p>
	 * 
	 * @param irods				The IRODS account to use.
	 * @param uploadDirectory	The directory in which all uploaded files must be stored.
	 *
	 * @throws IOException 				If there is an error while looking into the home directory of the given IRODS account.
	 * @throws NullPointerException 	If the given IRODS account is <i>null</i>.
	 * @throws UWSException				If the given root directory is not a directory or has not the READ and WRITE permissions.
	 * 
	 * @see iRodsUWSFileManager#iRodsUWSFileManager(IRODSAccount)
	 */
	public iRodsTAPFileManager(IRODSAccount irods, final File uploadDirectory) throws NullPointerException, IOException, UWSException {
		super(irods);
		this.uploadDirectory = uploadDirectory;
	}

	/**
	 * <p>Builds a {@link TAPFileManager} which manages all UWS files in the home directory of the given IRODS account.</p>
	 * <p>
	 * 	There will be one directory for each owner ID and owner directories will be grouped
	 * 	thanks to {@link DefaultOwnerGroupIdentifier}.
	 * </p>
	 * 
	 * @param irods				The IRODS account to use.
	 * @param logDirectory		The local directory in which all log files must be written. <i><u>note:</u> If NULL, all log files will be saved on IRODS.</i></p>
	 * @param uploadDirectory	The directory in which all uploaded files must be stored.
	 *
	 * @throws IOException 				If there is an error while looking into the home directory of the given IRODS account.
	 * @throws NullPointerException 	If the given IRODS account is <i>null</i>.
	 * @throws UWSException				If the given root directory is not a directory or has not the READ and WRITE permissions.
	 * 
	 * @see iRodsUWSFileManager#iRodsUWSFileManager(IRODSAccount, File))
	 */
	public iRodsTAPFileManager(IRODSAccount irods, final File logDirectory, final File uploadDirectory) throws NullPointerException, IOException, UWSException {
		super(irods, logDirectory);
		this.uploadDirectory = uploadDirectory;
	}

	/**
	 * <p>Builds a {@link TAPFileManager} which manages all UWS files in the given directory.</p>
	 * <p>
	 * 	There will be one directory for each owner ID and owner directories will be grouped
	 * 	thanks to {@link DefaultOwnerGroupIdentifier}.
	 * </p>
	 * 
	 * @param irods				The IRODS account to use.
	 * @param rootPath			UWS root directory.
	 * @param uploadDirectory	The directory in which all uploaded files must be stored.
	 *
	 * @throws IOException 				If there is an error while looking into the home directory of the given IRODS account.
	 * @throws NullPointerException 	If the given IRODS account is <i>null</i>.
	 * @throws UWSException				If the given root directory is not a directory or has not the READ and WRITE permissions.
	 * 
	 * @see iRodsUWSFileManager#iRodsUWSFileManager(IRODSAccount, String)
	 */
	public iRodsTAPFileManager(IRODSAccount irods, String rootPath, final File uploadDirectory) throws NullPointerException, IOException, UWSException {
		super(irods, rootPath);
		this.uploadDirectory = uploadDirectory;
	}

	/**
	 * <p>Builds a {@link TAPFileManager} which manages all UWS files in the given directory.</p>
	 * <p>
	 * 	There will be one directory for each owner ID and owner directories will be grouped
	 * 	thanks to {@link DefaultOwnerGroupIdentifier}.
	 * </p>
	 * 
	 * @param irods				The IRODS account to use.
	 * @param rootPath			UWS root directory.
	 * @param logDirectory		The local directory in which all log files must be written. <i><u>note:</u> If NULL, all log files will be saved on IRODS.</i></p>
	 * @param uploadDirectory	The directory in which all uploaded files must be stored.
	 *
	 * @throws IOException 				If there is an error while looking into the home directory of the given IRODS account.
	 * @throws NullPointerException 	If the given IRODS account is <i>null</i>.
	 * @throws UWSException				If the given root directory is not a directory or has not the READ and WRITE permissions.
	 * 
	 * @see iRodsUWSFileManager#iRodsUWSFileManager(IRODSAccount, String, File)
	 */
	public iRodsTAPFileManager(IRODSAccount irods, String rootPath, final File logDirectory, final File uploadDirectory) throws NullPointerException, IOException, UWSException {
		super(irods, rootPath, logDirectory);
		this.uploadDirectory = uploadDirectory;
	}

	/**
	 * <p>Builds a {@link TAPFileManager} which manages all UWS files in the given directory.</p>
	 * <p>
	 * 	If, according to the third parameter, the owner directories must be grouped,
	 * 	the {@link DefaultOwnerGroupIdentifier} will be used.
	 * </p>
	 * 
	 * @param irods						The IRODS account to use.
	 * @param rootPath					Path of the UWS root directory.
	 * @param oneDirectoryForEachUser	<i>true</i> to create one directory for each owner ID, <i>false</i> otherwise.
	 * @param groupUserDirectories		<i>true</i> to group user directories, <i>false</i> otherwise.
	 * 									<i><u>note:</u> this value is ignored if the previous parameter is false.</i>
	 * @param uploadDirectory			The directory in which all uploaded files must be stored.
	 *
	 * @throws IOException 				If there is an error while looking into the home directory of the given IRODS account.
	 * @throws NullPointerException 	If the given IRODS account is <i>null</i>.
	 * @throws UWSException				If the given root directory is not a directory or has not the READ and WRITE permissions.
	 * 
	 * @see iRodsUWSFileManager#iRodsUWSFileManager(IRODSAccount, String, boolean, boolean)
	 */
	public iRodsTAPFileManager(IRODSAccount irods, String rootPath, boolean oneDirectoryForEachUser, boolean groupUserDirectories, final File uploadDirectory) throws NullPointerException, IOException, UWSException {
		super(irods, rootPath, oneDirectoryForEachUser, groupUserDirectories);
		this.uploadDirectory = uploadDirectory;
	}

	/**
	 * <p>Builds a {@link TAPFileManager} which manages all UWS files in the given directory.</p>
	 * <p>
	 * 	If, according to the third parameter, the owner directories must be grouped,
	 * 	the {@link DefaultOwnerGroupIdentifier} will be used.
	 * </p>
	 * 
	 * @param irods						The IRODS account to use.
	 * @param rootPath					Path of the UWS root directory.
	 * @param oneDirectoryForEachUser	<i>true</i> to create one directory for each owner ID, <i>false</i> otherwise.
	 * @param groupUserDirectories		<i>true</i> to group user directories, <i>false</i> otherwise.
	 * 									<i><u>note:</u> this value is ignored if the previous parameter is false.</i>
	 * @param logDirectory				The local directory in which all log files must be written. <i><u>note:</u> If NULL, all log files will be saved on IRODS.</i></p>
	 * @param uploadDirectory			The directory in which all uploaded files must be stored.
	 *
	 * @throws IOException 				If there is an error while looking into the home directory of the given IRODS account.
	 * @throws NullPointerException 	If the given IRODS account is <i>null</i>.
	 * @throws UWSException				If the given root directory is not a directory or has not the READ and WRITE permissions.
	 * 
	 * @see iRodsUWSFileManager#iRodsUWSFileManager(IRODSAccount, String, boolean, boolean, File)
	 */
	public iRodsTAPFileManager(IRODSAccount irods, String rootPath, boolean oneDirectoryForEachUser, boolean groupUserDirectories, final File uploadDirectory, final File logDirectory) throws NullPointerException, IOException, UWSException {
		super(irods, rootPath, oneDirectoryForEachUser, groupUserDirectories, logDirectory);
		this.uploadDirectory = uploadDirectory;
	}

	/**
	 * Builds a {@link UWSFileManager} which manages all UWS files in the given directory.
	 * 
	 * @param irods						The IRODS account to use.
	 * @param rootPath					Path of the UWS root directory.
	 * @param oneDirectoryForEachUser	<i>true</i> to create one directory for each owner ID, <i>false</i> otherwise.
	 * @param groupUserDirectories		<i>true</i> to group user directories, <i>false</i> otherwise.
	 * 									<i><u>note:</u> this value is ignored if the previous parameter is false.</i>
	 * @param ownerGroupIdentifier		The "function" to use to identify the group of a job owner.
	 * 									<i><ul>
	 * 										<li><u>note 1:</u> this value is ignored if one of the two previous parameters is false.</li>
	 * 										<li><u>note 2:</u> if this value is null but the previous parameters are true,
	 * 											{@link DefaultOwnerGroupIdentifier} will be chosen as default group identifier.</li>
	 *									</ul></i>
	 * @param uploadDirectory			The directory in which all uploaded files must be stored.
	 *
	 * @throws IOException 				If there is an error while looking into the home directory of the given IRODS account.
	 * @throws NullPointerException 	If the given IRODS account is <i>null</i>.
	 * @throws UWSException				If the given root directory is not a directory or has not the READ and WRITE permissions.
	 * 
	 * @see iRodsUWSFileManager#iRodsUWSFileManager(IRODSAccount, String, boolean, boolean, OwnerGroupIdentifier)
	 */
	public iRodsTAPFileManager(IRODSAccount irods, String rootPath, boolean oneDirectoryForEachUser, boolean groupUserDirectories, OwnerGroupIdentifier ownerGroupIdentifier, final File uploadDirectory) throws NullPointerException, IOException, UWSException {
		super(irods, rootPath, oneDirectoryForEachUser, groupUserDirectories, ownerGroupIdentifier);
		this.uploadDirectory = uploadDirectory;
	}

	/**
	 * Builds a {@link UWSFileManager} which manages all UWS files in the given directory.
	 * 
	 * @param irods						The IRODS account to use.
	 * @param rootPath					Path of the UWS root directory.
	 * @param oneDirectoryForEachUser	<i>true</i> to create one directory for each owner ID, <i>false</i> otherwise.
	 * @param groupUserDirectories		<i>true</i> to group user directories, <i>false</i> otherwise.
	 * 									<i><u>note:</u> this value is ignored if the previous parameter is false.</i>
	 * @param ownerGroupIdentifier		The "function" to use to identify the group of a job owner.
	 * 									<i><ul>
	 * 										<li><u>note 1:</u> this value is ignored if one of the two previous parameters is false.</li>
	 * 										<li><u>note 2:</u> if this value is null but the previous parameters are true,
	 * 											{@link DefaultOwnerGroupIdentifier} will be chosen as default group identifier.</li>
	 *									</ul></i>
	 * @param logDirectory				The local directory in which all log files must be written. <i><u>note:</u> If NULL, all log files will be saved on IRODS.</i></p>
	 * @param uploadDirectory			The directory in which all uploaded files must be stored.
	 *
	 * @throws IOException 				If there is an error while looking into the home directory of the given IRODS account.
	 * @throws NullPointerException 	If the given IRODS account is <i>null</i>.
	 * @throws UWSException				If the given root directory is not a directory or has not the READ and WRITE permissions.
	 * 
	 * @see iRodsUWSFileManager#iRodsUWSFileManager(IRODSAccount, String, boolean, boolean, OwnerGroupIdentifier, File)
	 */
	public iRodsTAPFileManager(IRODSAccount irods, String rootPath, boolean oneDirectoryForEachUser, boolean groupUserDirectories, OwnerGroupIdentifier ownerGroupIdentifier, final File logDirectory, final File uploadDirectory) throws NullPointerException, IOException, UWSException {
		super(irods, rootPath, oneDirectoryForEachUser, groupUserDirectories, ownerGroupIdentifier, logDirectory);
		this.uploadDirectory = uploadDirectory;
	}

	@Override
	protected String getLogFileName(final String logTypeGroup) {
		if (logTypeGroup != null && logTypeGroup.equals(DBConnection.LOG_TYPE_DB_ACTIVITY.getCustomType()))
			return DEFAULT_DB_ACTIVITY_LOG_FILE_NAME;
		else
			return super.getLogFileName(logTypeGroup);
	}

	@Override
	public final File getUploadDirectory() {
		if (uploadDirectory != null && !uploadDirectory.exists())
			uploadDirectory.mkdirs();
		return uploadDirectory;
	}

}
