package uws.service.file;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.PrintWriter;

import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.NoSuchElementException;

import edu.sdsc.grid.io.GeneralFile;

import edu.sdsc.grid.io.irods.IRODSAccount;
import edu.sdsc.grid.io.irods.IRODSFile;
import edu.sdsc.grid.io.irods.IRODSFileInputStream;
import edu.sdsc.grid.io.irods.IRODSFileOutputStream;
import edu.sdsc.grid.io.irods.IRODSFileSystem;

import uws.UWSException;
import uws.UWSToolBox;

import uws.job.ErrorSummary;
import uws.job.Result;
import uws.job.UWSJob;
import uws.job.user.JobOwner;
import uws.service.log.UWSLogType;

/**
 * <p>All UWS files are stored in the local machine into the specified directory.</p>
 * <p>
 * 	The name of the log file, the result files and the backup files may be customized by overriding the following functions:
 * 	{@link #getLogFileName()}, {@link #getResultFileName(Result, AbstractJob)}, {@link #getBackupFileName(JobOwner)} and {@link #getBackupFileName()}.
 * </p>
 * <p>
 * 	By default, results and backups are grouped by owner/user and owners/users are grouped thanks to {@link DefaultOwnerGroupIdentifier}.
 * 	By using the appropriate constructor, you can change these default behaviors.
 * </p>
 * 
 * @author Gr&eacute;gory Mantelet (CDS)
 * @version 06/2012
 */
public class iRodsUWSFileManager implements UWSFileManager {

	/** Format to use to format dates. */
	private DateFormat dateFormat = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss.SSS");

	protected static final String DEFAULT_HTTP_LOG_FILE_NAME = "service_http_activity.log";
	protected static final String DEFAULT_DEBUG_LOG_FILE_NAME = "service_debug.log";
	protected static final String DEFAULT_LOG_FILE_NAME = "service_activity.log";
	protected static final String DEFAULT_BACKUP_FILE_NAME = "uws.backup";

	protected static final String UNKNOWN_LOG_TYPE_GROUP = "???";

	protected final IRODSAccount account;
	protected final IRODSFile rootDirectory;

	protected final boolean oneDirectoryForEachUser;
	protected final boolean groupUserDirectories;
	protected final OwnerGroupIdentifier ownerGroupId;

	protected final File logDirectory;
	protected Map<String, PrintWriter> logOutputs = new HashMap<String, PrintWriter>();


	/**
	 * <p>Builds a {@link UWSFileManager} which manages all UWS files in the home directory of the given IRODS account.</p>
	 * <p>
	 * 	There will be one directory for each owner ID and owner directories will be grouped
	 * 	thanks to {@link DefaultOwnerGroupIdentifier}.
	 * </p>
	 * 
	 * @param irods		The IRODS account to use.
	 *
	 * @throws IOException 				If there is an error while looking into the home directory of the given IRODS account.
	 * @throws NullPointerException 	If the given IRODS account is <i>null</i>.
	 * @throws UWSException				If the given root directory is not a directory or has not the READ and WRITE permissions.
	 * 
	 * @see #iRodsUWSFileManager(IRODSAccount, String, boolean, boolean, OwnerGroupIdentifier, File)
	 */
	public iRodsUWSFileManager(final IRODSAccount irods) throws NullPointerException, IOException, UWSException {
		this(irods, null, true, true, null, null);
	}

	/**
	 * <p>Builds a {@link UWSFileManager} which manages all UWS files in the home directory of the given IRODS account.</p>
	 * <p>
	 * 	There will be one directory for each owner ID and owner directories will be grouped
	 * 	thanks to {@link DefaultOwnerGroupIdentifier}.
	 * </p>
	 * 
	 * @param irods			The IRODS account to use.
	 * @param logDirectory	The local directory in which all log files must be written. <i><u>note:</u> If NULL, all log files will be saved on IRODS.</i></p>
	 *
	 * @throws IOException 				If there is an error while looking into the home directory of the given IRODS account.
	 * @throws NullPointerException 	If the given IRODS account is <i>null</i>.
	 * @throws UWSException				If the given root directory is not a directory or has not the READ and WRITE permissions.
	 * 
	 * @see #iRodsUWSFileManager(IRODSAccount, String, boolean, boolean, OwnerGroupIdentifier, File)
	 */
	public iRodsUWSFileManager(final IRODSAccount irods, final File logDirectory) throws NullPointerException, IOException, UWSException {
		this(irods, null, true, true, null, logDirectory);
	}

	/**
	 * <p>Builds a {@link UWSFileManager} which manages all UWS files in the given directory.</p>
	 * <p>
	 * 	There will be one directory for each owner ID and owner directories will be grouped
	 * 	thanks to {@link DefaultOwnerGroupIdentifier}.
	 * </p>
	 * 
	 * @param irods		The IRODS account to use.
	 * @param rootPath	UWS root directory.
	 *
	 * @throws IOException 				If there is an error while looking into the home directory of the given IRODS account.
	 * @throws NullPointerException 	If the given IRODS account is <i>null</i>.
	 * @throws UWSException				If the given root directory is not a directory or has not the READ and WRITE permissions.
	 * 
	 * @see #iRodsUWSFileManager(IRODSAccount, String, boolean, boolean, OwnerGroupIdentifier, File)
	 */
	public iRodsUWSFileManager(final IRODSAccount irods, final String rootPath) throws NullPointerException, IOException, UWSException {
		this(irods, rootPath, true, true, null, null);
	}

	/**
	 * <p>Builds a {@link UWSFileManager} which manages all UWS files in the given directory.</p>
	 * <p>
	 * 	There will be one directory for each owner ID and owner directories will be grouped
	 * 	thanks to {@link DefaultOwnerGroupIdentifier}.
	 * </p>
	 * 
	 * @param irods			The IRODS account to use.
	 * @param rootPath		UWS root directory.
	 * @param logDirectory	The local directory in which all log files must be written. <i><u>note:</u> If NULL, all log files will be saved on IRODS.</i></p>
	 *
	 * @throws IOException 				If there is an error while looking into the home directory of the given IRODS account.
	 * @throws NullPointerException 	If the given IRODS account is <i>null</i>.
	 * @throws UWSException				If the given root directory is not a directory or has not the READ and WRITE permissions.
	 * 
	 * @see #iRodsUWSFileManager(IRODSAccount, String, boolean, boolean, OwnerGroupIdentifier, File)
	 */
	public iRodsUWSFileManager(final IRODSAccount irods, final String rootPath, final File logDirectory) throws NullPointerException, IOException, UWSException {
		this(irods, rootPath, true, true, null, logDirectory);
	}

	/**
	 * <p>Builds a {@link UWSFileManager} which manages all UWS files in the given directory.</p>
	 * <p>
	 * 	If, according to the third parameter, the owner directories must be grouped,
	 * 	the {@link DefaultOwnerGroupIdentifier} will be used.
	 * </p>
	 * 
	 * @param irods						The IRODS account to use.
	 * @param rootPath					Path of the UWS root directory.
	 * @param oneDirectoryForEachUser	<i>true</i> to create one directory for each owner ID, <i>false</i> otherwise.
	 * @param groupUserDirectories		<i>true</i> to group user directories, <i>false</i> otherwise.
	 * 									<i><u>note:</u> this value is ignored if the previous parameter is false.</i>
	 *
	 * @throws IOException 				If there is an error while looking into the home directory of the given IRODS account.
	 * @throws NullPointerException 	If the given IRODS account is <i>null</i>.
	 * @throws UWSException				If the given root directory is not a directory or has not the READ and WRITE permissions.
	 * 
	 * @see #iRodsUWSFileManager(IRODSAccount, String, boolean, boolean, OwnerGroupIdentifier, File)
	 */
	public iRodsUWSFileManager(final IRODSAccount irods, final String rootPath, final boolean oneDirectoryForEachUser, final boolean groupUserDirectories) throws NullPointerException, IOException, UWSException {
		this(irods, rootPath, oneDirectoryForEachUser, groupUserDirectories, null, null);
	}

	/**
	 * <p>Builds a {@link UWSFileManager} which manages all UWS files in the given directory.</p>
	 * <p>
	 * 	If, according to the third parameter, the owner directories must be grouped,
	 * 	the {@link DefaultOwnerGroupIdentifier} will be used.
	 * </p>
	 * 
	 * @param irods						The IRODS account to use.
	 * @param rootPath					Path of the UWS root directory.
	 * @param oneDirectoryForEachUser	<i>true</i> to create one directory for each owner ID, <i>false</i> otherwise.
	 * @param groupUserDirectories		<i>true</i> to group user directories, <i>false</i> otherwise.
	 * 									<i><u>note:</u> this value is ignored if the previous parameter is false.</i>
	 * @param logDirectory				The local directory in which all log files must be written. <i><u>note:</u> If NULL, all log files will be saved on IRODS.</i></p>
	 *
	 * @throws IOException 				If there is an error while looking into the home directory of the given IRODS account.
	 * @throws NullPointerException 	If the given IRODS account is <i>null</i>.
	 * @throws UWSException				If the given root directory is not a directory or has not the READ and WRITE permissions.
	 * 
	 * @see #iRodsUWSFileManager(IRODSAccount, String, boolean, boolean, OwnerGroupIdentifier, File)
	 */
	public iRodsUWSFileManager(final IRODSAccount irods, final String rootPath, final boolean oneDirectoryForEachUser, final boolean groupUserDirectories, final File logDirectory) throws NullPointerException, IOException, UWSException {
		this(irods, rootPath, oneDirectoryForEachUser, groupUserDirectories, null, logDirectory);
	}

	/**
	 * Builds a {@link UWSFileManager} which manages all UWS files in the given directory.
	 * 
	 * @param irods						The IRODS account to use.
	 * @param rootPath					Path of the UWS root directory.
	 * @param oneDirectoryForEachUser	<i>true</i> to create one directory for each owner ID, <i>false</i> otherwise.
	 * @param groupUserDirectories		<i>true</i> to group user directories, <i>false</i> otherwise.
	 * 									<i><u>note:</u> this value is ignored if the previous parameter is false.</i>
	 * @param ownerGroupIdentifier		The "function" to use to identify the group of a job owner.
	 * 									<i><ul>
	 * 										<li><u>note 1:</u> this value is ignored if one of the two previous parameters is false.</li>
	 * 										<li><u>note 2:</u> if this value is null but the previous parameters are true,
	 * 											{@link DefaultOwnerGroupIdentifier} will be chosen as default group identifier.</li>
	 *									</ul></i>
	 *
	 * @throws IOException 				If there is an error while looking into the home directory of the given IRODS account.
	 * @throws NullPointerException 	If the given IRODS account is <i>null</i>.
	 * @throws UWSException				If the given root directory is not a directory or has not the READ and WRITE permissions.
	 * 
	 * @see #iRodsUWSFileManager(IRODSAccount, String, boolean, boolean, OwnerGroupIdentifier, File)
	 */
	public iRodsUWSFileManager(final IRODSAccount irods, final String rootPath, final boolean oneDirectoryForEachUser, final boolean groupUserDirectories, final OwnerGroupIdentifier ownerGroupIdentifier) throws NullPointerException, IOException, UWSException {
		this(irods, rootPath, oneDirectoryForEachUser, groupUserDirectories, ownerGroupIdentifier, null);
	}

	/**
	 * Builds a {@link UWSFileManager} which manages all UWS files in the given directory.
	 * 
	 * @param irods						The IRODS account to use.
	 * @param rootPath					Path of the UWS root directory.
	 * @param oneDirectoryForEachUser	<i>true</i> to create one directory for each owner ID, <i>false</i> otherwise.
	 * @param groupUserDirectories		<i>true</i> to group user directories, <i>false</i> otherwise.
	 * 									<i><u>note:</u> this value is ignored if the previous parameter is false.</i>
	 * @param ownerGroupIdentifier		The "function" to use to identify the group of a job owner.
	 * 									<i><ul>
	 * 										<li><u>note 1:</u> this value is ignored if one of the two previous parameters is false.</li>
	 * 										<li><u>note 2:</u> if this value is null but the previous parameters are true,
	 * 											{@link DefaultOwnerGroupIdentifier} will be chosen as default group identifier.</li>
	 *									</ul></i>
	 * @param logDirectory				The local directory in which all log files must be written. <i><u>note:</u> If NULL, all log files will be saved on IRODS.</i></p>
	 *
	 * @throws IOException 				If there is an error while looking into the home directory of the given IRODS account.
	 * @throws NullPointerException 	If the given IRODS account is <i>null</i>.
	 * @throws UWSException				If the given root directory is not a directory or has not the READ and WRITE permissions.
	 */
	public iRodsUWSFileManager(final IRODSAccount irods, final String rootPath, final boolean oneDirectoryForEachUser, final boolean groupUserDirectories, final OwnerGroupIdentifier ownerGroupIdentifier, final File logDirectory) throws NullPointerException, IOException, UWSException {
		account = irods;

		IRODSFileSystem fileSystem = new IRODSFileSystem(account);
		if (rootPath == null || rootPath.trim().isEmpty())
			rootDirectory = new IRODSFile(fileSystem, ".");
		else
			rootDirectory = new IRODSFile(fileSystem, rootPath);

		if (rootDirectory == null)
			throw new NullPointerException("Missing root directory ! Impossible to create a LocalUWSFileManager.");
		else if (!rootDirectory.exists()){
			if (!rootDirectory.mkdirs())
				throw new UWSException(UWSException.INTERNAL_SERVER_ERROR, "The given root directory does not exist and can not be created automatically !");
		}else if (!rootDirectory.isDirectory())
			throw new UWSException(UWSException.INTERNAL_SERVER_ERROR, "The root directory of a UWSFileManager must be a DIRECTORY !");
		else if (!rootDirectory.canRead())
			throw new UWSException(UWSException.INTERNAL_SERVER_ERROR, "Missing READ permission for the root directory of a UWSFileManager !");
		else if (!rootDirectory.canWrite())
			throw new UWSException(UWSException.INTERNAL_SERVER_ERROR, "Missing WRITE permission for the root directory of a UWSFileManager !");

		this.oneDirectoryForEachUser = oneDirectoryForEachUser;
		if (this.oneDirectoryForEachUser){
			this.groupUserDirectories = groupUserDirectories;
			if (this.groupUserDirectories){
				if (ownerGroupIdentifier != null)
					this.ownerGroupId = ownerGroupIdentifier;
				else
					this.ownerGroupId = new DefaultOwnerGroupIdentifier();
			}else
				this.ownerGroupId = null;
		}else{
			this.groupUserDirectories = false;
			this.ownerGroupId = null;
		}

		if (logDirectory != null){
			if (!logDirectory.isDirectory())
				throw new UWSException(UWSException.INTERNAL_SERVER_ERROR, "The given log directory \""+logDirectory.getAbsolutePath()+"\" is not a directory !");
			if (!logDirectory.canRead())
				throw new UWSException(UWSException.INTERNAL_SERVER_ERROR, "Missing READ permission for the given log directory \""+logDirectory.getAbsolutePath()+"\" !");
			if (!logDirectory.canWrite())
				throw new UWSException(UWSException.INTERNAL_SERVER_ERROR, "Missing WRITE permission for the given log directory \""+logDirectory.getAbsolutePath()+"\" !");
			else
				this.logDirectory = logDirectory;
		}else
			this.logDirectory = null;
	}

	/**
	 * Gets the directory of the given owner.
	 * 
	 * @param owner	A job owner.
	 * @return		Its directory.
	 */
	public IRODSFile getOwnerDirectory(final JobOwner owner){
		if (!oneDirectoryForEachUser || owner == null || owner.getID() == null || owner.getID().trim().isEmpty())
			return rootDirectory;

		IRODSFile ownerDir = rootDirectory;
		if (groupUserDirectories){
			String ownerGroup = ownerGroupId.getOwnerGroup(owner);
			if (ownerGroup != null)
				ownerDir = new IRODSFile(rootDirectory, ownerGroup);
		}
		ownerDir = new IRODSFile(ownerDir, owner.getID().replaceAll(File.separator, "_"));

		return ownerDir;
	}

	/**
	 * Removes the owner directory if there is no more file in it (except the backup file which is no more required).
	 * 
	 * @param owner			The user whose the directory must be removed.
	 * 
	 * @throws IOException	If there is an error while removing the owner directory.
	 */
	protected void cleanOwnerDirectory(final JobOwner owner) throws IOException {
		// Remove the owner directory if empty or if only the owner backup file exists:
		if (owner != null && oneDirectoryForEachUser){
			IRODSFile ownerDir = getOwnerDirectory(owner);
			String[] dirContent = ownerDir.list();
			// if empty...
			if (dirContent.length <= 1){
				// delete the owner backup file it it exists:
				if (dirContent.length == 1 && dirContent[0].equals(getBackupFileName(owner))){
					(new IRODSFile(rootDirectory, getBackupFileName(owner))).delete();
					dirContent = ownerDir.list();
				}
				// if empty (so, if the owner backup file does not exist any more)...
				if (dirContent.length == 0){
					// delete the owner directory !
					if (ownerDir.delete() && groupUserDirectories){
						// if the user group directory is also empty now, delete it:
						GeneralFile userGroupDir = ownerDir.getParentFile();
						if (userGroupDir.list().length == 0)
							userGroupDir.delete();
					}
				}
			}
		}
	}

	/* ******************* */
	/* LOG FILE MANAGEMENT */
	/* ******************* */
	/**
	 * <p>Lets grouping log messages by log type.</p>
	 * <p>For instance: by default all messages of type INFO, WARNING and ERROR are written in the same file.</p>
	 * 
	 * @param logType	Type of the message to log.
	 * 
	 * @return			Name of the log type group.
	 */
	protected String getLogTypeGroup(final UWSLogType logType) {
		switch(logType){
		case INFO:
		case WARNING:
		case ERROR:
			return "DefaultLog";
		case DEBUG:
		case HTTP_ACTIVITY:
			return logType.toString();
		case CUSTOM:
			return logType.getCustomType();
		default:
			return UNKNOWN_LOG_TYPE_GROUP;
		}
	}

	/**
	 * <p>Gets the name of the UWS log file.</p>
	 * <p>By default: {@link #DEFAULT_LOG_FILE_NAME} or {@link #DEFAULT_HTTP_LOG_FILE_NAME} (to log an activity message, that's to say: thread status or http request).</p>
	 * 
	 * @param logTypeGroup	Type of message to log.
	 * 
	 * @return	The name of the UWS log file.
	 */
	protected String getLogFileName(final String logTypeGroup){
		if (logTypeGroup == UWSLogType.HTTP_ACTIVITY.toString())
			return DEFAULT_HTTP_LOG_FILE_NAME;
		else if (logTypeGroup.equals(UWSLogType.DEBUG.toString()))
			return DEFAULT_DEBUG_LOG_FILE_NAME;
		else
			return DEFAULT_LOG_FILE_NAME;
	}

	/**
	 * Gets the UWS log file on IRODS.
	 * 
	 * @param logType	Type of message to log.
	 * 
	 * @return	The UWS log file on IRODS.
	 * 
	 * @see #getLogFileName()
	 */
	protected IRODSFile getIRODSLogFile(final String logTypeGroup){
		return new IRODSFile(rootDirectory, getLogFileName(logTypeGroup));
	}

	/**
	 * Gets the UWS log file on the local machine.
	 * 
	 * @param logTypeString	Type of message to log.
	 * 
	 * @return	The UWS log file on the local machine.
	 * 
	 * @see #getLogFileName()
	 */
	protected File getLocalLogFile(final String logTypeGroup){
		if (logDirectory != null)
			return new File(logDirectory, getLogFileName(logTypeGroup));
		else
			return null;
	}

	@Override
	public InputStream getLogInput(final UWSLogType logType) throws IOException {
		String logTypeGroup = getLogTypeGroup(logType);
		if (logDirectory == null){
			IRODSFile logFile = getIRODSLogFile(logTypeGroup);
			if (logFile.exists())
				return new IRODSFileInputStream(logFile);
			else
				return null;
		}else{
			File logFile = getLocalLogFile(logTypeGroup);
			if (logFile.exists())
				return new FileInputStream(logFile);
			else
				return null;
		}
	}

	@Override
	public PrintWriter getLogOutput(final UWSLogType logType) throws IOException {
		String logTypeGroup = getLogTypeGroup(logType);
		PrintWriter output = logOutputs.get(logTypeGroup);
		if (output == null){
			if (logDirectory == null){
				IRODSFile logFile = getIRODSLogFile(logTypeGroup);
				createParentDir(logFile);
				output = new PrintWriter(new IRODSFileOutputStream(logFile), true);
			}else{
				File logFile = getLocalLogFile(logTypeGroup);
				createParentDir(logFile);
				output = new PrintWriter(new FileOutputStream(logFile, true), true);
			}
			printLogHeader(output);
			logOutputs.put(logTypeGroup, output);
		}
		return output;
	}

	/**
	 * Print a header into the log file so that separating older log messages to the new ones.
	 */
	protected void printLogHeader(final PrintWriter out){
		String msgHeader = "########################################### LOG STARTS "+dateFormat.format(new Date())+" ###########################################";
		StringBuffer buf = new StringBuffer("");
		for(int i=0; i<msgHeader.length(); i++)
			buf.append('#');
		String separator = buf.toString();

		out.println(separator);
		out.println(msgHeader);
		out.println(separator);

		out.flush();
	}

	/* *********************** */
	/* RESULT FILES MANAGEMENT */
	/* *********************** */
	/**
	 * <p>Gets the name of the file in which the given result is/must be written.</p>
	 * <p>By default: jobID + "_" + resultID + "." + {@link UWSToolBox#getFileExtension(String) getFileExtension(resultMIMEType)}</p>
	 * <p><i><u>note:</u> there is no file extension if the MIME type of the result is unknown !</i></p>
	 * 
	 * @param result	The result whose the file name is asked.
	 * @param job		The job which owns the given result.
	 * 
	 * @return			Name of the file corresponding to the given result.
	 * 
	 * @see UWSToolBox#getFileExtension(String)
	 */
	protected String getResultFileName(final Result result, final UWSJob job){
		String fileName = job.getJobId()+"_";

		if (result != null && result.getId() != null && !result.getId().trim().isEmpty())
			fileName += result.getId();
		else
			fileName += Result.DEFAULT_RESULT_NAME;

		String fileExt = UWSToolBox.getFileExtension(result.getMimeType());
		fileExt = (fileExt == null) ? "" : ("."+fileExt);
		fileName += fileExt;

		return fileName;
	}

	/**
	 * Gets the file corresponding to the given result.
	 * 
	 * @param result	The result whose the file is asked.
	 * @param job		The job which owns the given result.
	 * 
	 * @return			The file corresponding to the given result.
	 * 
	 * @see #getOwnerDirectory(JobOwner)
	 * @see #getResultFileName(Result, AbstractJob)
	 */
	protected IRODSFile getResultFile(final Result result, final UWSJob job){
		IRODSFile ownerDir = getOwnerDirectory(job.getOwner());
		return new IRODSFile(ownerDir, getResultFileName(result, job));
	}

	@Override
	public InputStream getResultInput(Result result, UWSJob job) throws IOException {
		IRODSFile resultFile = getResultFile(result, job);
		return resultFile.exists() ? new IRODSFileInputStream(resultFile) : null;
	}

	@Override
	public OutputStream getResultOutput(Result result, UWSJob job) throws IOException {
		IRODSFile resultFile = getResultFile(result, job);
		createParentDir(resultFile);
		return new IRODSFileOutputStream(resultFile);
	}

	@Override
	public long getResultSize(Result result, UWSJob job) throws IOException {
		IRODSFile resultFile = getResultFile(result, job);
		if (resultFile == null || !resultFile.exists())
			return -1;
		else
			return resultFile.length();
	}

	@Override
	public boolean deleteResult(Result result, UWSJob job) throws IOException {
		boolean deleted = getResultFile(result, job).delete();

		if (deleted)
			cleanOwnerDirectory(job.getOwner());

		return deleted;
	}


	/* ********************** */
	/* ERROR FILES MANAGEMENT */
	/* ********************** */
	/**
	 * <p>Gets the name of the file in which the described error is/must be written.</p>
	 * <p>By default: jobID + "_ERROR.log"</p>
	 * 
	 * @param error		The description of the error whose the file name is asked.
	 * @param job		The job which owns the given error.
	 * 
	 * @return			Name of the file corresponding to the described error.
	 */
	protected String getErrorFileName(final ErrorSummary error, final UWSJob job) {
		return job.getJobId()+"_ERROR.log";
	}

	/**
	 * Gets the file corresponding to the described error.
	 * 
	 * @param error		The error whose the file is asked.
	 * @param job		The job which owns the given error.
	 * 
	 * @return			The file corresponding to the described error.
	 * 
	 * @see #getOwnerDirectory(JobOwner)
	 * @see #getErrorFileName(ErrorSummary, UWSJob)
	 */
	protected IRODSFile getErrorFile(final ErrorSummary error, final UWSJob job){
		IRODSFile ownerDir = getOwnerDirectory(job.getOwner());
		return new IRODSFile(ownerDir, getErrorFileName(error, job));
	}

	@Override
	public InputStream getErrorInput(ErrorSummary error, UWSJob job) throws IOException {
		IRODSFile resultFile = getErrorFile(error, job);
		return resultFile.exists() ? new IRODSFileInputStream(resultFile) : null;
	}

	@Override
	public OutputStream getErrorOutput(ErrorSummary error, UWSJob job) throws IOException {
		IRODSFile resultFile = getErrorFile(error, job);
		createParentDir(resultFile);
		return new IRODSFileOutputStream(resultFile);
	}

	@Override
	public long getErrorSize(ErrorSummary error, UWSJob job) throws IOException {
		IRODSFile errorFile = getErrorFile(error, job);
		if (errorFile == null || !errorFile.exists())
			return -1;
		else
			return errorFile.length();
	}

	@Override
	public boolean deleteError(ErrorSummary error, UWSJob job) throws IOException {
		boolean deleted = getErrorFile(error, job).delete();

		if (deleted)
			cleanOwnerDirectory(job.getOwner());

		return deleted;
	}

	/* *********************** */
	/* BACKUP FILES MANAGEMENT */
	/* *********************** */
	/**
	 * <p>Gets the name of the backup file of the given job owner (~ UWS user).</p>
	 * <p>By default: ownerID + ".backup"</p>
	 * 
	 * @param owner	The job owner whose the name of the backup file is asked.
	 * 
	 * @return		The name of the backup file of the given owner.
	 * 
	 * @throws IllegalArgumentException	If the given owner is <i>null</i> or an empty string.
	 */
	protected String getBackupFileName(final JobOwner owner) throws IllegalArgumentException {
		if (owner == null || owner.getID() == null || owner.getID().trim().isEmpty())
			throw new IllegalArgumentException("Missing owner ! Can not get the backup file of an unknown owner. See LocalUWSFileManager.getBackupFile(JobOwner)");
		return owner.getID().replaceAll(File.separator, "_")+".backup";
	}

	@Override
	public Iterator<InputStream> getAllUserBackupInputs() {
		return new IRODSAllUserBackupInputs(this);
	}

	@Override
	public InputStream getBackupInput(JobOwner owner) throws IllegalArgumentException, IOException {
		IRODSFile backupFile = new IRODSFile(getOwnerDirectory(owner), getBackupFileName(owner));
		return backupFile.exists() ? new IRODSFileInputStream(backupFile) : null;
	}

	@Override
	public OutputStream getBackupOutput(JobOwner owner) throws IllegalArgumentException, IOException {
		IRODSFile backupFile = new IRODSFile(getOwnerDirectory(owner), getBackupFileName(owner));
		createParentDir(backupFile);
		return new IRODSFileOutputStream(backupFile);
	}

	/**
	 * <p>Gets the name of the UWS general backup file.</p>
	 * <p>By default: {@link #DEFAULT_BACKUP_FILE_NAME}</p>
	 * 
	 * @return		The name of the UWS general backup file.
	 */
	protected String getBackupFileName() {
		return DEFAULT_BACKUP_FILE_NAME;
	}

	@Override
	public InputStream getBackupInput() throws IOException {
		IRODSFile backupFile = new IRODSFile(rootDirectory, getBackupFileName());
		return backupFile.exists() ? new IRODSFileInputStream(backupFile) : null;
	}

	@Override
	public OutputStream getBackupOutput() throws IOException {
		IRODSFile backupFile = new IRODSFile(rootDirectory, getBackupFileName());
		createParentDir(backupFile);
		return new IRODSFileOutputStream(backupFile);
	}


	/* ************** */
	/* TOOL FUNCTIONS */
	/* ************** */

	/**
	 * Creates the parent directory(ies) if it(they) does/do not exist.
	 * 
	 * @param f	The file whose the parent directory must exist after the call of this function.
	 * 
	 * @return	<i>true</i> if the parent directory now exists, <i>false</i> otherwise.
	 */
	protected boolean createParentDir(final IRODSFile f){
		if (!f.getParentFile().exists())
			return f.getParentFile().mkdirs();
		else
			return true;
	}

	/**
	 * Creates the parent directory(ies) if it(they) does/do not exist.
	 * 
	 * @param f	The file whose the parent directory must exist after the call of this function.
	 * 
	 * @return	<i>true</i> if the parent directory now exists, <i>false</i> otherwise.
	 */
	protected boolean createParentDir(final File f){
		if (!f.getParentFile().exists())
			return f.getParentFile().mkdirs();
		else
			return true;
	}

	/**
	 * Lets iterating on all user backup files.
	 * The {@link #next()} function creates and returns the {@link InputStream} for the next backup file.
	 * 
	 * @author Gr&eacute;gory Mantelet (CDS)
	 * @version 05/2012
	 */
	protected class IRODSAllUserBackupInputs implements Iterator<InputStream> {

		private final iRodsUWSFileManager fileManager;
		private Iterator<IRODSFile> itBackupFiles;

		public IRODSAllUserBackupInputs(final iRodsUWSFileManager fm){
			fileManager = fm;
			itBackupFiles = loadAllBackupFiles().iterator();
		}

		private ArrayList<IRODSFile> loadAllBackupFiles(){
			ArrayList<IRODSFile> backupFiles = new ArrayList<IRODSFile>();

			// If there must be 1 directory by user:
			if (fileManager.oneDirectoryForEachUser){
				ArrayList<IRODSFile> dir0 = listDirectories(fileManager.rootDirectory);
				// If user directories must be grouped (so, we have the list of all user groups):
				if (fileManager.groupUserDirectories){
					for(IRODSFile groupDir : dir0){
						ArrayList<IRODSFile> dir1 = listDirectories(groupDir);
						for(IRODSFile userDir : dir1)
							addOwnerBackupFiles(backupFiles, userDir, userDir.getName());
					}
				}// Otherwise: We have already all user directories:
				else{
					for(IRODSFile userDir : dir0)
						addOwnerBackupFiles(backupFiles, userDir, userDir.getName());
				}

			}// Otherwise: Get all backup files into the root directory:
			else
				addOwnerBackupFiles(backupFiles, fileManager.rootDirectory, null);

			return backupFiles;
		}

		private ArrayList<IRODSFile> listDirectories(IRODSFile rootDir){
			ArrayList<IRODSFile> directories = new ArrayList<IRODSFile>();
			String[] fileNames = rootDir.list();
			for(String fileName : fileNames){
				IRODSFile f = new IRODSFile(rootDir, fileName);
				if (f.isDirectory())
					directories.add(f);
			}
			return directories;
		}

		private void addOwnerBackupFiles(final ArrayList<IRODSFile> files, final IRODSFile rootDirectory, final String ownerID){
			String[] backups = rootDirectory.list();
			for(String fileName : backups){
				if (fileName == null || fileName.trim().isEmpty())
					continue;

				if (ownerID != null && !ownerID.trim().isEmpty()){
					if (!fileName.equalsIgnoreCase(ownerID+".backup"))
						continue;
				}else{
					if (!fileName.endsWith(".backup") || fileName.equalsIgnoreCase(getBackupFileName()))
						continue;
				}

				IRODSFile f = new IRODSFile(rootDirectory, fileName);
				if (!f.isDirectory())
					files.add(f);
			}
		}

		@Override
		public boolean hasNext() {
			return itBackupFiles != null && itBackupFiles.hasNext();
		}

		/**
		 * If the file whose the input stream must be created and returned does not exist
		 * or has not the READ permission, <i>null</i> will be returned.
		 * 
		 * @see java.util.Iterator#next()
		 */
		@Override
		public InputStream next() throws NoSuchElementException {
			if (itBackupFiles == null)
				throw new NoSuchElementException();

			try {
				IRODSFile f = itBackupFiles.next();
				if (!itBackupFiles.hasNext())
					itBackupFiles = null;

				return (f == null || !f.exists()) ? null : new IRODSFileInputStream(f);

			} catch (IOException e) {
				// TODO Log the error !
				return null;
			}
		}

		@Override
		public void remove() {
			throw new UnsupportedOperationException();
		}

	}

}
